@@ -1,59 +1,72 @@
-from curses import tigetnum, tigetstr, setupterm, tparm
-import inspect
+from curses import tigetstr, setupterm, tparm
+from fcntl import ioctl
+from functools import partial
 from itertools import cycle
-import logging
 import os
+from signal import signal, SIGWINCH
+import struct
+from termios import TIOCGWINSZ
 from traceback import format_list, extract_tb, format_exception_only
-from unittest import TestCase
 
 from nose.plugins import Plugin
 from nose.exc import SkipTest
-from nose.case import FunctionTestCase
 from nose.util import test_address
 
 
-log = logging.getLogger('nose.plugins.ProgressivePlugin')
-
-
 class ProgressivePlugin(Plugin):
     """Nose plugin which prioritizes the important information"""
     name = 'progressive'
-    testsRun = 0
+    _testsRun = 0
+    _totalTests = 0
 
     # TODO: Decrease score so we run early, and monkeypatch stderr in __init__.
     # See if that works.
 
-    def options(self, parser, env=os.environ):
-        super(ProgressivePlugin, self).options(parser, env=env)
-        self.parser = parser
-
-    def configure(self, options, conf):
-        super(ProgressivePlugin, self).configure(options, conf)
-        if self.enabled:
-            self.cmd_options = options
-            self.config = conf
-
     def setOutputStream(self, stream):
         """Steal the stream, and return a mock one for everybody else to shut them up."""
         class DummyStream(object):
             writeln = flush = write = lambda self, *args: None
 
-        self.stream = stream
+        # 1 in case test counting failed and returned 0
+        self.bar = ProgressBar(stream, self._totalTests or 1)
+
+        self.stream = self.bar.stream = stream
         # Explicit args make setupterm() work even when -s is passed:
-        setupterm(None, stream.fileno())
+        setupterm(None, stream.fileno())  # so things like tigetstr() work
         # TODO: Don't call setupterm() if self.stream isn't a terminal. Use
         # os.isatty(self.stream.fileno()). If it isn't, perhaps replace the
         # ShyProgressBar with a dummy object.
 
         return DummyStream()
 
-    def prepareTest(self, test):
-        """Init the progress bar, and tell it how many tests there are."""
-        # Not really clear from the docs whether this always gets called in
-        # every plugin or whether another plugin can prevent this from getting
-        # called by returning something. If the latter, we'll be surprised when
-        # self.bar doesn't exist.
-        self.bar = ProgressBar(test.countTestCases())
+    def prepareTestLoader(self, loader):
+        """Insert ourselves into loader calls to count tests.
+
+        The top-level loader call often returns lazy results, like a LazySuite.
+        This is a problem, as we would destroy the suite by iterating over it
+        to count the tests. Consequently, we monkeypatch the top-level loader
+        call to do the load twice: once for the actual test running and again
+        to yield something we can iterate over to do the count.
+
+        """
+        def capture_suite(orig_method, *args, **kwargs):
+            """Intercept calls to the loader before they get lazy.
+
+            Re-execute them to grab a copy of the possibly lazy suite, and
+            count the tests therein.
+
+            """
+            self._totalTests += orig_method(*args, **kwargs).countTestCases()
+            return orig_method(*args, **kwargs)
+
+        # TODO: If there's ever a practical need, also patch loader.suiteClass
+        # or even TestProgram.createTests. createTests seems to be main top-
+        # level caller of loader methods, and nose.core.collector() (which
+        # isn't even called in nose) is an alternate one.
+        if hasattr(loader, 'loadTestsFromNames'):
+            loader.loadTestsFromNames = partial(capture_suite,
+                                                loader.loadTestsFromNames)
+        return loader
 
     def printError(self, kind, err, test):
         """Output a human-readable error report to the stream.
@@ -71,12 +84,12 @@ class ProgressivePlugin(Plugin):
         # TODO: Canonicalize the path to remove /kitsune/../kitsune nonsense.
         # Don't relativize, though, as that hurts the ability to paste into
         # running editors.
-        with ShyProgressBar(self.stream, self.bar):
+        with self.bar.dodging():
             writeln = self.stream.writeln
             write = self.stream.write
             writeln()
             write(tigetstr('bold'))
-            writeln('%s: %s' % (kind, python_path(test)))
+            writeln('%s: %s' % (kind, nose_selector(test)))
 
             # File name and line num in a format vi can take:
             address = test.address()
@@ -95,16 +108,15 @@ class ProgressivePlugin(Plugin):
             # Exception:
             self.stream.write(''.join(format_exception_only(exception_type, exception_value)))
 
-
     def printErrors(self):
         # The current summary doesn't begin with a \n.
-        with ShyProgressBar(self.stream, self.bar):
+        with self.bar.dodging():
             self.stream.writeln()
 
     def finalize(self, result):
         """Print counts of tests run."""
         types = ['test', 'failure', 'error']
-        values = [self.testsRun, len(result.failures), len(result.errors)]
+        values = [self._testsRun, len(result.failures), len(result.errors)]
         if hasattr(result, 'skipped'):  # Absent if --no-skip is passed
             types.append('skip')
             values.append(len(result.skipped))
@@ -113,20 +125,18 @@ class ProgressivePlugin(Plugin):
 
         # Erase progress bar. Bash doesn't clear the whole line when printing
         # the prompt, leaving a piece of the bar.
-        with AtProgressBar(self.stream):
-            self.stream.write(tigetstr('el'))
+        self.bar.erase()
         self.stream.writeln('\n' + msg)
 
     def startTest(self, test):
         # Overriding this seems to prevent TextTestRunner from running its, so
         # we have to keep track of the test count ourselves.
-        self.testsRun += 1
-        with AtProgressBar(self.stream):
-            self.stream.write(self.bar.get(test, self.testsRun))
+        self._testsRun += 1
+        self.bar.update(test, self._testsRun)
 
     def addError(self, test, err):
         exc, val, tb = err
-        with ShyProgressBar(self.stream, self.bar):
+        with self.bar.dodging():
             if isinstance(exc, SkipTest):
                 self.stream.writeln()
                 self.stream.writeln('SKIP: %s' % nose_selector(test))
@@ -136,25 +146,30 @@ class ProgressivePlugin(Plugin):
     def addFailure(self, test, err):
         self.printError('FAIL', err, test)
 
-    def addSkip(self, test, reason):
-        # Only in 2.7+
-        # TODO: This never gets called, at least in 2.6.
-        with ShyProgressBar(self.stream, self.bar):
-            self.stream.writeln()
-            self.stream.writeln('SKIP: %s' % nose_selector(test))
-
 
 class ProgressBar(object):
     SPINNER_CHARS = r'/-\|'
 
-    def __init__(self, max):
+    def __init__(self, stream, max):
         """max is the highest value I will attain. Must be >0."""
+        self.stream = stream
         self.last = ''  # The contents of the previous progress line printed
         self.max = max
-        self.spinner = cycle(self.SPINNER_CHARS)
+        self._spinner = cycle(self.SPINNER_CHARS)
+        self._measure_terminal()
+        signal(SIGWINCH, self._handle_winch)
 
-    def get(self, test, number):
-        """Return updated content for the progress bar.
+    def _measure_terminal(self):
+        self.lines, self.cols = \
+            struct.unpack('hhhh', ioctl(0, TIOCGWINSZ, '\000' * 8))[0:2]
+
+    def _handle_winch(self, *args):
+        #self.erase()  # Doesn't seem to help.
+        self._measure_terminal()
+        # TODO: Reprint the bar but at the new width.
+
+    def update(self, test, number):
+        """Draw an updated progress bar.
 
         At the moment, the graph takes a fixed width, and the test identifier
         takes the rest of the row, truncated from the left to fit.
@@ -163,88 +178,80 @@ class ProgressBar(object):
         number -- how many tests have been run so far, including this one
 
         """
-        # TODO: Play nicely with absurdly narrow terminals.
+        # TODO: Play nicely with absurdly narrow terminals. (OS X's won't even
+        # go small enough to hurt us.)
 
         # Figure out graph:
-        # We cheat a bit and dedicate extra column to the spinner to make the
-        # logic simple, so the graph is just sliiiightly misproportional.
         GRAPH_WIDTH = 14
-        num_markers = int(round(float(number) / self.max * GRAPH_WIDTH))
+        # min() is in case we somehow get the total test count wrong. It's tricky.
+        num_markers = int(round(min(1.0, float(number) / self.max) * GRAPH_WIDTH))
         # If there are any markers, replace the last one with the spinner.
         # Otherwise, have just a spinner:
-        markers = '=' * (num_markers - 1) + self.spinner.next()
-        graph = '[%s%s]' % (markers,
-                            ' ' * (GRAPH_WIDTH - len(markers)))
+        markers = '=' * (num_markers - 1) + self._spinner.next()
+        graph = '[%s%s]' % (markers, ' ' * (GRAPH_WIDTH - len(markers)))
 
         # Figure out the test identifier portion:
-        test_path = python_path(test.test) + '.' + test.test._testMethodName
-        width = tigetnum('cols')
-        cols_for_path = width - len(graph) - 2  # 2 spaces between path & graph
+        test_path = nose_selector(test)
+        cols_for_path = self.cols - len(graph) - 2  # 2 spaces between path & graph
         if len(test_path) > cols_for_path:
             test_path = test_path[len(test_path) - cols_for_path:]
         else:
             test_path += ' ' * (cols_for_path - len(test_path))
-        
+
         # Put them together, and let simmer:
         self.last = tigetstr('bold') + test_path + '  ' + graph + tigetstr('sgr0')
-        return self.last
+        with AtLine(self.stream, self.lines):
+            self.stream.write(self.last)
+
+    def erase(self):
+        """White out the progress bar."""
+        with AtLine(self.stream, self.lines):
+            self.stream.write(tigetstr('el'))
+        self.stream.flush()
+
+    def dodging(bar):
+        """Return a context manager which erases the bar, lets you output things, and then redraws the bar."""
+        class ShyProgressBar(object):
+            """Context manager that implements a progress bar that gets out of the way"""
+
+            def __enter__(self):
+                """Erase the progress bar so bits of disembodied progress bar don't get scrolled up the terminal."""
+                # My terminal has no status line, so we make one manually.
+                bar.erase()
+
+            def __exit__(self, type, value, tb):
+                """Redraw the last saved state of the progress bar."""
+                # This isn't really necessary unless we monkeypatch stderr; the
+                # next test is about to start and will redraw the bar.
+                with AtLine(bar.stream, bar.lines):
+                    bar.stream.write(bar.last)
+                bar.stream.flush()
+
+        return ShyProgressBar()
 
 
-class AtProgressBar(object):
-    """Context manager which goes to the progress bar line on entry and goes back to where it was on exit"""
+class AtLine(object):
+    """Context manager which moves the cursor to a certain line on entry and goes back to where it was on exit"""
 
-    def __init__(self, stream):
+    def __init__(self, stream, line):
         self.stream = stream
+        self.line = line
 
     def __enter__(self):
         """Save position and move to progress bar, col 1."""
         self.stream.write(tigetstr('sc'))  # save position
-        height = tigetnum('lines')
-        self.stream.write(tparm(tigetstr('cup'), height, 0))
+        self.stream.write(tparm(tigetstr('cup'), self.line, 0))
 
     def __exit__(self, type, value, tb):
         self.stream.write(tigetstr('rc'))  # restore position
 
 
-class ShyProgressBar(object):
-    """Context manager that implements a progress bar that gets out of the way"""
-
-    def __init__(self, stream, bar):
-        self.stream = stream
-        self.bar = bar
-
-    def __enter__(self):
-        """Erase the progress bar so bits of disembodied progress bar don't get scrolled up the terminal."""
-        # My terminal has no status line, so we make one manually.
-        with AtProgressBar(self.stream):
-            self.stream.write(tigetstr('el'))  # erase to EOL
-        self.stream.flush()
-
-    def __exit__(self, type, value, tb):
-        """Do nothing; the bar will come back at the top of the next test, which is imminent."""
-        # This isn't really necessary unless we monkeypatch stderr; the next
-        # test is about to start and will redraw the bar.
-        with AtProgressBar(self.stream):
-            self.stream.write(self.bar.last)
-        self.stream.flush()
-
-
 def nose_selector(test):
     """Return the string you can pass to nose to run `test`."""
-    file, module, rest = test_address(test)
-    file = relative_source_path(file)
-    if rest:
-        return '%s:%s' % (file, rest)
-    else:
-        return file
-
-
-def python_path(test):
     address = test_address(test)
-    if address:
-        file, module, rest = address
-    else:
+    if not address:
         return 'Unknown test'
+    file, module, rest = address
 
     if rest:
         return '%s:%s' % (module, rest)
